@@ -5,24 +5,8 @@ import pprint
 
 from . import core
 
-try:
-    import cPickle as pickle
-except ImportError:
-    import pickle
+import pickle
 
-import sys
-if sys.version_info[0] == 2:
-    range = xrange
-    string_types = basestring
-elif sys.version_info[0] == 3:
-    string_types = str
-
-
-def string_output(s):
-    if sys.version_info[0] == 2:
-        return s
-    elif sys.version_info[0] == 3:
-        return s.decode('UTF-8')
 
 RT_Memory = 0
 RT_Disk = 1
@@ -61,10 +45,12 @@ def _get_bounds(handle, bounds_fn, interleaved):
     if (dimension.value == 0):
         return None
 
-    mins = ctypes.cast(pp_mins, ctypes.POINTER(ctypes.c_double
-                                               * dimension.value))
-    maxs = ctypes.cast(pp_maxs, ctypes.POINTER(ctypes.c_double
-                                               * dimension.value))
+    mins = ctypes.cast(
+        pp_mins, ctypes.POINTER(ctypes.c_double * dimension.value)
+    )
+    maxs = ctypes.cast(
+        pp_maxs, ctypes.POINTER(ctypes.c_double * dimension.value)
+    )
 
     results = [mins.contents[i] for i in range(dimension.value)]
     results += [maxs.contents[i] for i in range(dimension.value)]
@@ -94,7 +80,7 @@ def _get_data(handle):
 class Index(object):
     """An R-Tree, MVR-Tree, or TPR-Tree indexing object"""
 
-    def __init__(self,  *args, **kwargs):
+    def __init__(self, *args, **kwargs):
         """Creates a new index
 
         :param filename:
@@ -160,8 +146,12 @@ class Index(object):
             >>> p = index.Property()
 
             >>> idx = index.Index(properties=p)
-            >>> idx  # doctest: +ELLIPSIS
-            <rtree.index.Index object at 0x...>
+            >>> idx  # doctest: +NORMALIZE_WHITESPACE
+            rtree.index.Index(bounds=[1.7976931348623157e+308,
+                                    1.7976931348623157e+308,
+                                    -1.7976931348623157e+308,
+                                    -1.7976931348623157e+308],
+                                    size=0)
 
         Insert an item into the index::
 
@@ -220,7 +210,7 @@ class Index(object):
         basename = None
         storage = None
         if args:
-            if isinstance(args[0], string_types) or isinstance(args[0], bytes):
+            if isinstance(args[0], str) or isinstance(args[0], bytes):
                 # they sent in a filename
                 basename = args[0]
                 # they sent in a filename, stream
@@ -283,7 +273,6 @@ class Index(object):
         else:
             self.properties.storage = RT_Memory
 
-
         ps = kwargs.get('pagesize', None)
         if ps:
             self.properties.pagesize = int(ps)
@@ -299,6 +288,16 @@ class Index(object):
                 for item in stream:
                     self.insert(*item)
 
+    def get_size(self):
+        try:
+            return self.count(self.bounds)
+        except core.RTreeError:
+            return 0
+
+    def __repr__(self):
+        return 'rtree.index.Index(bounds={}, size={})'.format(self.bounds,
+                                                              self.get_size())
+
     def __getstate__(self):
         state = self.__dict__.copy()
         del state["handle"]
@@ -386,7 +385,6 @@ class Index(object):
         # return serialized to keep it alive for the pointer.
         return size, ctypes.cast(p, ctypes.POINTER(ctypes.c_uint8)), serialized
 
-
     def set_result_limit(self, value):
         return core.rt.Index_SetResultSetOffset(self.handle, value)
 
@@ -463,7 +461,7 @@ class Index(object):
         p_mins, p_maxs = self.get_coordinate_pointers(coordinates)
         pv_mins, pv_maxs = self.get_coordinate_pointers(velocities)
         # End time isn't used
-        t_start, t_end = self._get_time_doubles((time, time+1))
+        t_start, t_end = self._get_time_doubles((time, time + 1))
         data = ctypes.c_ubyte(0)
         size = 0
         if obj is not None:
@@ -529,7 +527,6 @@ class Index(object):
 
         return p_num_results.value
 
-
     def _countTP(self, coordinates, velocities, times):
         p_mins, p_maxs = self.get_coordinate_pointers(coordinates)
         pv_mins, pv_maxs = self.get_coordinate_pointers(velocities)
@@ -777,7 +774,7 @@ class Index(object):
                         yield self.loads(data)
 
             core.rt.Index_DestroyObjResults(its, num_results)
-        except:  # need to catch all exceptions, not just rtree.
+        except Exception:  # need to catch all exceptions, not just rtree.
             core.rt.Index_DestroyObjResults(its, num_results)
             raise
 
@@ -790,7 +787,7 @@ class Index(object):
             for i in range(num_results):
                 yield items.contents[i]
             core.rt.Index_Free(its)
-        except:
+        except Exception:
             core.rt.Index_Free(its)
             raise
 
@@ -854,6 +851,14 @@ class Index(object):
             return self._nearest_obj(coordinates, num_results, objects)
         p_mins, p_maxs = self.get_coordinate_pointers(coordinates)
 
+        # p_num_results is an input and output for C++ lib
+        # as an input it says "get n closest neighbors"
+        # but if multiple neighbors are at the same distance, both
+        # will be returned
+        # so the number of returned neighbors may be > p_num_results
+        # thus p_num_results.contents.value gets set as an output by the
+        # C++ lib to indicate the actual number of results for
+        # _get_ids to use
         p_num_results = ctypes.pointer(ctypes.c_uint64(num_results))
 
         it = ctypes.pointer(ctypes.c_int64())
@@ -865,10 +870,14 @@ class Index(object):
                                           ctypes.byref(it),
                                           p_num_results)
 
-        return self._get_ids(it, min(num_results,p_num_results.contents.value))
+        return self._get_ids(it, p_num_results.contents.value)
 
-    def _nearestTP(self, coordinates, velocities, times, num_results=1,
-                  objects=False):
+    def _nearestTP(self,
+                   coordinates,
+                   velocities,
+                   times,
+                   num_results=1,
+                   objects=False):
         p_mins, p_maxs = self.get_coordinate_pointers(coordinates)
         pv_mins, pv_maxs = self.get_coordinate_pointers(velocities)
         t_start, t_end = self._get_time_doubles(times)
@@ -907,13 +916,17 @@ class Index(object):
     bounds = property(get_bounds)
 
     def delete(self, id, coordinates):
-        """Deletes items from the index with the given ``'id'`` within the
-        specified coordinates.
+        """Deletes an item from the index with the given ``'id'`` and
+           coordinates given by the ``coordinates`` sequence. As the index can
+           contain multiple items with the same ID and coordinates, deletion
+           is not guaranteed to delete all items in the index with the given ID
+           and coordinates.
 
         :param id: long integer
-            A long integer that is the identifier for this index entry.  IDs
-            need not be unique to be inserted into the index, and it is up
-            to the user to ensure they are unique if this is a requirement.
+            A long integer ID for the entry, which need not be unique. The
+            index can contain multiple entries with identical IDs and
+            coordinates. Uniqueness of items should be enforced at the
+            application level by the user.
 
         :param coordinates: sequence or array
             Dimension * 2 coordinate pairs, representing the min
@@ -1045,8 +1058,8 @@ class Index(object):
             # this code assumes the coords are not interleaved.
             # xmin, xmax, ymin, ymax, zmin, zmax
             for i in range(dimension):
-                mins[i] = coordinates[i*2]
-                maxs[i] = coordinates[(i*2)+1]
+                mins[i] = coordinates[i * 2]
+                maxs[i] = coordinates[(i * 2) + 1]
 
             p_mins[0] = ctypes.cast(mins, ctypes.POINTER(ctypes.c_double))
             p_maxs[0] = ctypes.cast(maxs, ctypes.POINTER(ctypes.c_double))
@@ -1140,6 +1153,7 @@ class Index(object):
 
         return output
 
+
 # An alias to preserve backward compatibility
 Rtree = Index
 
@@ -1204,8 +1218,8 @@ class Handle(object):
         try:
 
             if self._ptr is not None:
-               self._destroy(self._ptr)
-               self._ptr = None
+                self._destroy(self._ptr)
+                self._ptr = None
         except AttributeError:
             pass
 
@@ -1235,10 +1249,11 @@ class IndexHandle(Handle):
         try:
             core.rt.Index_Flush
             if self._ptr is not None:
-               core.rt.Index_Flush(self._ptr)
+                core.rt.Index_Flush(self._ptr)
         except AttributeError:
             pass
 
+
 class IndexStreamHandle(IndexHandle):
 
     _create = core.rt.Index_CreateWithStream
@@ -1515,11 +1530,10 @@ class Property(object):
     """Reinsert factor"""
 
     def get_filename(self):
-        s = core.rt.IndexProperty_GetFileName(self.handle)
-        return string_output(s)
+        return core.rt.IndexProperty_GetFileName(self.handle).decode()
 
     def set_filename(self, value):
-        if isinstance(value, string_types):
+        if isinstance(value, str):
             value = value.encode('utf-8')
         return core.rt.IndexProperty_SetFileName(self.handle, value)
 
@@ -1527,11 +1541,11 @@ class Property(object):
     """Index filename for disk storage"""
 
     def get_dat_extension(self):
-        s = core.rt.IndexProperty_GetFileNameExtensionDat(self.handle)
-        return string_output(s)
+        ext = core.rt.IndexProperty_GetFileNameExtensionDat(self.handle)
+        return ext.decode()
 
     def set_dat_extension(self, value):
-        if isinstance(value, string_types):
+        if isinstance(value, str):
             value = value.encode('utf-8')
         return core.rt.IndexProperty_SetFileNameExtensionDat(
             self.handle, value)
@@ -1540,11 +1554,11 @@ class Property(object):
     """Extension for .dat file"""
 
     def get_idx_extension(self):
-        s = core.rt.IndexProperty_GetFileNameExtensionIdx(self.handle)
-        return string_output(s)
+        ext = core.rt.IndexProperty_GetFileNameExtensionIdx(self.handle)
+        return ext.decode()
 
     def set_idx_extension(self, value):
-        if isinstance(value, string_types):
+        if isinstance(value, str):
             value = value.encode('utf-8')
         return core.rt.IndexProperty_SetFileNameExtensionIdx(
             self.handle, value)
@@ -1835,8 +1849,12 @@ class RtreeContainer(Rtree):
             >>> p = index.Property()
 
             >>> idx = index.RtreeContainer(properties=p)
-            >>> idx  # doctest: +ELLIPSIS
-            <rtree.index.RtreeContainer object at 0x...>
+            >>> idx  # doctest: +NORMALIZE_WHITESPACE
+            rtree.index.RtreeContainer(bounds=[1.7976931348623157e+308,
+                                     1.7976931348623157e+308,
+                                     -1.7976931348623157e+308,
+                                     -1.7976931348623157e+308],
+                                     size=0)
 
         Insert an item into the index::
 
@@ -1854,7 +1872,7 @@ class RtreeContainer(Rtree):
             [34.37768294..., 26.73758537..., 49.37768294..., 41.73758537...]
         """
         if args:
-            if isinstance(args[0], string_types) \
+            if isinstance(args[0], str) \
                     or isinstance(args[0], bytes) \
                     or isinstance(args[0], ICustomStorage):
                 raise ValueError('%s supports only in-memory indexes'
@@ -1862,6 +1880,16 @@ class RtreeContainer(Rtree):
         self._objects = {}
         return super(RtreeContainer, self).__init__(*args, **kwargs)
 
+    def get_size(self):
+        try:
+            return self.count(self.bounds)
+        except core.RTreeError:
+            return 0
+
+    def __repr__(self):
+        m = 'rtree.index.RtreeContainer(bounds={}, size={})'
+        return m.format(self.bounds, self.get_size())
+
     def __contains__(self, obj):
         return id(obj) in self._objects
 
@@ -1974,11 +2002,11 @@ class RtreeContainer(Rtree):
             49.3776829412, 41.7375853734])]
 
         """
-        if bbox == False:
+        if bbox is False:
             for id in super(RtreeContainer,
                             self).intersection(coordinates, bbox):
                 yield self._objects[id][1]
-        elif bbox == True:
+        elif bbox is True:
             for value in super(RtreeContainer,
                                self).intersection(coordinates, bbox):
                 value.object = self._objects[value.id][1]
@@ -1988,7 +2016,7 @@ class RtreeContainer(Rtree):
             raise ValueError(
                 "valid values for the bbox argument are True and False")
 
-    def nearest(self, coordinates, num_results = 1, bbox=False):
+    def nearest(self, coordinates, num_results=1, bbox=False):
         """Returns the ``k``-nearest objects to the given coordinates
         in increasing distance order.
 
@@ -2022,11 +2050,11 @@ class RtreeContainer(Rtree):
             >>> idx.insert(object(), (34.37, 26.73, 49.37, 41.73))
             >>> hits = idx.nearest((0, 0, 10, 10), 3, bbox=True)
         """
-        if bbox == False:
+        if bbox is False:
             for id in super(RtreeContainer,
                             self).nearest(coordinates, num_results, bbox):
                 yield self._objects[id][1]
-        elif bbox == True:
+        elif bbox is True:
             for value in super(RtreeContainer,
                                self).nearest(coordinates, num_results, bbox):
                 value.object = self._objects[value.id][1]
